
#include "Arduino.h"
#define SET_ACQUISITION 100
#define STOP_ACQUISITION 101
#define SET_FILTRAGE_16BITS 102
#define SET_FILTRAGE_32BITS 103
#define SET_FILTRAGE_FLOAT 104
#define NBUF 0x4
#define NBUF_MASK 0x3 // (NBUF-1)
#define BUF_SIZE 0x100
#define BUF_MASK 0xFF // (BUF_SIZE-1)
volatile uint8_t buf[NBUF][BUF_SIZE]; // tampon pour la transmission
volatile uint8_t compteur_buf,compteur_buf_transmis;
volatile uint16_t indice_buf;
uint16_t diviseur[6] = {0,1,8,64,256,1024};
uint32_t nblocs;
uint8_t sans_fin;
uint8_t flag;
uint8_t reduc;
uint8_t i_reduc;

void (*isrCallback)();

#define MAX_NCOEF 50
uint8_t nb; // nombre de coefficients b
int16_t b[MAX_NCOEF]; // coefficients b
int32_t bb[MAX_NCOEF];
float bbb[MAX_NCOEF];
uint8_t na; // nombre de coefficients a
int16_t a[MAX_NCOEF]; // coefficients a
int32_t aa[MAX_NCOEF];
float aaa[MAX_NCOEF];
uint8_t ab_shift;
uint8_t max_na_nb;
#define TAILLE_TAMPON 0x100 // supérieur à max(na,nb)
#define MASQUE_TAMPON 0xFF // TAILLE_TAMPON-1
int8_t x[TAILLE_TAMPON]; // tampon pour les x_k
int8_t y[TAILLE_TAMPON]; // tampon pour les y_k
int32_t ww[TAILLE_TAMPON];
float www[TAILLE_TAMPON];
int16_t v[MAX_NCOEF];
int32_t vv[MAX_NCOEF];


uint8_t J_tampon; // indice pour les tampons
uint8_t offset;
            
void timer1_init(uint32_t period) {
    TCCR1A = 0;
    TCCR1B = 0;
    TCCR1B |= (1 << WGM12);
    uint32_t top = (F_CPU/1000000*period);
    int clock = 1;
    while ((top>0xFFFF)&&(clock<5)) {
          clock++;
          top = (F_CPU/1000000*period/diviseur[clock]);
      }
    OCR1A = top;
    OCR1B = top >> 1;
    TIMSK1 = (1 << OCIE1B);
    TCCR1B |= clock;
    
}
              
void stop_acquisition() {
   TCCR1B &= ~0x7; 
}
              
void adc_init(uint8_t multiplex, uint8_t prescaler,void (*isr)()) {
  ADMUX = (1 << REFS0);
  ADMUX |= multiplex & 0b00011111;
  ADMUX |= (1 << ADLAR); //left adjust
  ADCSRA = 0;
  ADCSRA |= (1 << ADEN); // enable ADC
  ADCSRA |= (1 << ADATE); // Auto Trigger Enable
  ADCSRA |= (1 << ADIE); // interrupt enable
  ADCSRA |= prescaler ;
  ADCSRB = 0;
  ADCSRB |= (multiplex & 0b00100000) >> 2;
  ADCSRB |= (1 << ADTS2)|(1 << ADTS0);// trigger source : timer 1 comp B
  isrCallback = isr;
}
                
ISR(ADC_vect) {
     isrCallback();
}
                
ISR(TIMER1_COMPB_vect) {
    if (flag==0) {
         flag = 1;
         PORTD &= ~(1<<PORTD3); // sortie 18 sur Arduino MEGA, 3 sur UNO
       }
     else {
        flag = 0;
        PORTD |= (1<<PORTD3);   
     }   
}
                  
void adc() {
   if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = ADCH;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     } 
   }
}
                 
void adc_filtrage_16bits_FD1() {
    int16_t accum;
    int16_t sortie;
    uint8_t i,k;
    x[J_tampon] = ADCH-offset;
    accum = 0;
    i = J_tampon;
    for (k=0; k<nb; k++) {
      accum += x[i]*b[k];
      i--; 
    }
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= y[i]*a[k];
      i--;  
    }
    sortie = accum >> ab_shift;
    y[J_tampon] = sortie;
    J_tampon++;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
    }
}
                 
void adc_filtrage_16bits_FD2() {
    int16_t accum;
    int16_t sortie;
    uint8_t i,k;
    accum = (ADCH-offset)*a[0];
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= ww[i]*a[k];
      i--; 
    }
    ww[J_tampon] = accum >> ab_shift;
    i = J_tampon;
    accum = 0;
    for (k=0; k<nb; k++) {
      accum += ww[i]*b[k];
      i--;  
    }
    sortie = accum >> ab_shift;
    J_tampon++;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
    } 
}
                 
void adc_filtrage_16bits_FDT2() {
    int16_t entree;
    int16_t sortie;
    uint8_t k,i;
    i = 0;
    entree = ADCH-offset;
    sortie = (b[0]*entree+v[0]) >> ab_shift;
    for (k=1; k<max_na_nb; k++) {
      v[i] = entree*b[k] - sortie*a[k] + v[i+1];
      i++;  
    }
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
    }
}
                   
void adc_filtrage_32bits_FD1() {
    int32_t accum;
    int16_t sortie;
    uint8_t i,k;
    x[J_tampon] = ADCH-offset;
    accum = 0;
    i = J_tampon;
    for (k=0; k<nb; k++) {
      accum += x[i]*bb[k];
      i--; 
    }
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= y[i]*aa[k];
      i--;  
    }
    sortie = accum >> ab_shift;
    y[J_tampon] = sortie;
    J_tampon++;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
     }
}
                         
void adc_filtrage_32bits_FD2() {
    int32_t accum;
    int16_t sortie;
    uint8_t i,k;
    accum = (ADCH-offset)*aa[0];
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= ww[i]*aa[k];
      i--; 
    }
    ww[J_tampon] = accum >> ab_shift;
    i = J_tampon;
    accum = 0;
    for (k=0; k<nb; k++) {
      accum += ww[i]*bb[k];
      i--;  
    }
    J_tampon++;
    sortie = accum >> ab_shift;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] =  sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
     }
}
                            
void adc_filtrage_32bits_FDT2() {
    int16_t in;
    int16_t out;
    uint8_t k,i;
    i = 0;
    in = ADCH-offset;
    out = (bb[0]*in+vv[0]) >> ab_shift;
    for (k=1; k<max_na_nb; k++) {
      vv[i] = in*bb[k] - out*aa[k] + vv[i+1];
      i++;  
    }
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = out+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
    }
}
                           
void adc_filtrage_float_FD1() {
    float accum;
    uint16_t sortie;
    uint8_t i,k;
    x[J_tampon] = ADCH-offset;
    accum = 0;
    i = J_tampon;
    for (k=0; k<nb; k++) {
      accum += x[i]*bbb[k];
      i--; 
    }
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= y[i]*aaa[k];
      i--;  
    }
    sortie = accum;
    y[J_tampon] = sortie;
    J_tampon++;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
     }
}
                     
void adc_filtrage_float_FD2() {
    float accum;
    uint8_t i,k;
    accum = (ADCH-offset)*aaa[0];
    i = J_tampon-1;
    for (k=1; k<na; k++) {
      accum -= www[i]*aaa[k];
      i--; 
    }
    www[J_tampon] = accum;
    accum = 0;
    i = J_tampon;
    for (k=0; k<nb; k++) {
      accum += www[i]*bbb[k];
      i--;  
    }
    J_tampon++;
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = accum+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
     }
}
                     
void adc_filtrage_float_FDT2() {
    int16_t entree,sortie;
    uint8_t k;
    uint8_t i;
    i = 0;
    entree = ADCH-offset;
    sortie = bbb[0]*entree+www[0];
    for (k=1; k<max_na_nb; k++) {
      www[i] = bbb[k]*entree - aaa[k]*sortie + www[i+1];
      i++; 
    }
    if (--i_reduc == 0) {
    i_reduc = reduc;
    buf[compteur_buf][indice_buf] = sortie+offset;
    indice_buf++;
    if (indice_buf == BUF_SIZE) {
       indice_buf = 0;
       compteur_buf = (compteur_buf+1)&NBUF_MASK;
     }
    }
    
}
                     
void setup() {
  char c;
  Serial.begin(500000);
  Serial.setTimeout(0);
  c = 0;
  Serial.write(c);
  c = 255;
  Serial.write(c);
  c = 0;
  Serial.write(c);
  compteur_buf = compteur_buf_transmis = 0;
  indice_buf = 0;
  DDRD |= 1 << PORTD3; // PD3 en sortie
  flag = 0;
}
                        
void lecture_acquisition() {
  uint32_t c1,c2,c3,c4;
  uint8_t multiplex,prescaler;
  uint32_t period;
  while (Serial.available()<2) {};
  multiplex = Serial.read();
  prescaler = Serial.read();
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  period = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  nblocs = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<1);
  reduc = Serial.read();
  i_reduc = reduc;
  if (nblocs==0) {
        sans_fin = 1;
        nblocs = 1;
   }
   else sans_fin = 0;
   compteur_buf=compteur_buf_transmis=0;
   indice_buf = 0;
  cli();
  timer1_init(period);
  adc_init(multiplex,prescaler,adc);
  sei();
}
                          
void lecture_filtrage_16bits() {
  uint32_t c1,c2,c3,c4;
  uint8_t multiplex,prescaler;
  uint32_t period;
  uint8_t type;
  while (Serial.available()<2) {};
  multiplex = Serial.read();
  prescaler = Serial.read();
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  period = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  nblocs = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<1);
  reduc = Serial.read();
  i_reduc = reduc;
   // filtrage
   while (Serial.available()<1) {};
   na = Serial.read();
   for (int k=0; k<na; k++) {
       while (Serial.available()<2) {};
       c1 = Serial.read();
       c2 = Serial.read();
       a[k] = ((c1<<8)|c2);
   }
   while (Serial.available()<1) {};
   nb = Serial.read();
   for (int k=0; k<nb; k++) {
       while (Serial.available()<2) {};
       c1 = Serial.read();
       c2 = Serial.read();
       b[k] = ((c1<<8)|c2);
   }
   while (Serial.available()<1) {};
   ab_shift = Serial.read();
   while (Serial.available()<1) {};
   offset = Serial.read();
   while (Serial.available()<1) {};
   type = Serial.read();
   
   if (nblocs==0) {
        sans_fin = 1;
        nblocs = 1;
   }
   else sans_fin = 0;
   compteur_buf=compteur_buf_transmis=0;
   indice_buf = 0;
   for (int i=0; i<= TAILLE_TAMPON-1; i++) {
     x[i] = 0;
     y[i] = 0;
     ww[i] = 0;
   }
  J_tampon = 0;
  cli();
  timer1_init(period);
  if (type==1) adc_init(multiplex,prescaler, adc_filtrage_16bits_FD1);
  else if (type==2) adc_init(multiplex,prescaler, adc_filtrage_16bits_FD2);
  else if (type==3) {
    max_na_nb = max(na,nb);
    int k;
    for (k=na; k<max_na_nb; k++) a[k] = 0;
    for (k=nb; k<max_na_nb; k++) b[k] = 0;
    adc_init(multiplex,prescaler, adc_filtrage_16bits_FDT2);
  }
  sei();
}
                           
void lecture_filtrage_32bits() {
  uint32_t c1,c2,c3,c4;
  uint8_t multiplex,prescaler;
  uint32_t period;
  uint8_t type;
  while (Serial.available()<2) {};
  multiplex = Serial.read();
  prescaler = Serial.read();
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  period = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  nblocs = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<1);
  reduc = Serial.read();
  i_reduc = reduc;
   // filtrage
   while (Serial.available()<1) {};
   na = Serial.read();
   for (int k=0; k<na; k++) {
       while (Serial.available()<4) {};
       c1 = Serial.read();
       c2 = Serial.read();
       c3 = Serial.read();
       c4 = Serial.read();
       aa[k] = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
   }
   while (Serial.available()<1) {};
   nb = Serial.read();
   for (int k=0; k<nb; k++) {
       while (Serial.available()<4) {};
       c1 = Serial.read();
       c2 = Serial.read();
       c3 = Serial.read();
       c4 = Serial.read();
       bb[k] = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
   }
   while (Serial.available()<1) {};
   ab_shift = Serial.read();
   while (Serial.available()<1) {};
   offset = Serial.read();
   while (Serial.available()<1) {};
   type = Serial.read();
   
   if (nblocs==0) {
        sans_fin = 1;
        nblocs = 1;
       
   }
   else sans_fin = 0;
   compteur_buf=compteur_buf_transmis=0;
   indice_buf = 0;
   for (int i=0; i<= TAILLE_TAMPON-1; i++) {
     x[i] = 0;
     y[i] = 0;
     ww[i] = 0;
   }
  J_tampon = 0;
  cli();
  timer1_init(period);
  if (type==1) adc_init(multiplex,prescaler, adc_filtrage_32bits_FD1);
  else if (type==2) adc_init(multiplex,prescaler, adc_filtrage_32bits_FD2);
  else if (type==3) {
    max_na_nb = max(na,nb);
    int k;
    for (k=na; k<max_na_nb; k++) aa[k] = 0;
    for (k=nb; k<max_na_nb; k++) bb[k] = 0;
    adc_init(multiplex,prescaler, adc_filtrage_32bits_FDT2);
  }
  sei();
}

                            
void lecture_filtrage_float() {
  uint32_t c1,c2,c3,c4;
  int32_t m;
  int8_t e;
  uint8_t multiplex,prescaler;
  uint32_t period;
  uint8_t type;
  while (Serial.available()<2) {};
  multiplex = Serial.read();
  prescaler = Serial.read();
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  period = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<4) {};
  c1 = Serial.read();
  c2 = Serial.read();
  c3 = Serial.read();
  c4 = Serial.read();
  nblocs = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
  while (Serial.available()<1);
  reduc = Serial.read();
  i_reduc = reduc;
   // filtrage
   while (Serial.available()<1) {};
   na = Serial.read();
   for (int k=0; k<na; k++) {
       while (Serial.available()<5) {};
       c1 = Serial.read();
       c2 = Serial.read();
       c3 = Serial.read();
       c4 = Serial.read();
       e = Serial.read();
       m = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
       aaa[k] = ((float)m)*pow(2,e-30);
   }
   while (Serial.available()<1) {};
   nb = Serial.read();
   for (int k=0; k<nb; k++) {
       while (Serial.available()<5) {};
       c1 = Serial.read();
       c2 = Serial.read();
       c3 = Serial.read();
       c4 = Serial.read();
       e = Serial.read();
       m = ((c1 << 24) | (c2 << 16) | (c3 << 8) | c4);
       bbb[k] = ((float)m)*pow(2,e-30);
   }
   while (Serial.available()<1) {};
   offset = Serial.read();
   while (Serial.available()<1) {};
   type = Serial.read();
   
   if (nblocs==0) {
        sans_fin = 1;
        nblocs = 1;
   }
   else sans_fin = 0;
   compteur_buf=compteur_buf_transmis=0;
   indice_buf = 0;
   for (int i=0; i<= TAILLE_TAMPON-1; i++) {
     x[i] = 0;
     y[i] = 0;
     www[i] = 0.0;  
   }
  J_tampon = 0;
  cli();
  timer1_init(period);
  if (type==1) adc_init(multiplex,prescaler, adc_filtrage_float_FD1);
  else if (type==2) adc_init(multiplex,prescaler, adc_filtrage_float_FD2);
  else if (type==3) {
    max_na_nb = max(na,nb);
    int k;
    for (k=na; k<max_na_nb; k++) aaa[k] = 0.0;
    for (k=nb; k<max_na_nb; k++) bbb[k] = 0.0;
    adc_init(multiplex,prescaler, adc_filtrage_float_FDT2);
  }
  sei();
}
                             
void lecture_serie() {
   char com;
   if (Serial.available()>0) {
        com = Serial.read();
        if (com==SET_ACQUISITION) lecture_acquisition();
        if (com==STOP_ACQUISITION) stop_acquisition();
        if (com==SET_FILTRAGE_16BITS) lecture_filtrage_16bits();
        if (com==SET_FILTRAGE_32BITS) lecture_filtrage_32bits();
        if (com==SET_FILTRAGE_FLOAT) lecture_filtrage_float();
   }
}
                              
void loop() {
  if ((compteur_buf_transmis!=compteur_buf)&&(nblocs>0)) {
      Serial.write((uint8_t *)buf[compteur_buf_transmis],BUF_SIZE);
      compteur_buf_transmis=(compteur_buf_transmis+1)&NBUF_MASK;
      if (sans_fin==0) {
            nblocs--;
            if (nblocs==0) stop_acquisition();
      }
  }
  else lecture_serie();
}                         
                               